<?php
/**
 * Plugin Name: Easy Replace Image
 * Plugin URI: https://iuliacazan.ro/easy-replace-image/
 * Description: This plugin allows you to replace an attachment file by uploading another image or by downloading one from a specified URL, without deleting the attachment. The plugin handles the image sizes generation and the attachment metadata update.
 * Text Domain: srifu
 * Domain Path: /langs
 * Version: 2.1
 * Author: Iulia Cazan
 * Author URI: https://profiles.wordpress.org/iulia-cazan
 * Donate link: https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=JJA37EHZXWUTJ
 * License: GPL2
 *
 * @package srifu
 *
 * Copyright (C) 2017-2019 Iulia Cazan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

// Define the plugin version.
define( 'SERIFU_PLUGIN_VERSION', 2.1 );

/**
 * Class for Replace Image From Upload or URL
 */
class Sisanu_Easy_Replace_Image {

	const PLUGIN_NAME        = 'Easy Replace Image';
	const PLUGIN_SUPPORT_URL = 'https://wordpress.org/support/plugin/easy-replace-image/';
	const PLUGIN_TRANSIENT   = 'srifu-plugin-notice';

	/**
	 * Class instance.
	 *
	 * @var object
	 */
	private static $instance;

	/**
	 * Get active object instance
	 *
	 * @return object
	 */
	public static function get_instance() {
		if ( ! self::$instance ) {
			self::$instance = new Sisanu_Easy_Replace_Image();
		}
		return self::$instance;
	}

	/**
	 * Class constructor, includes and init method.
	 *
	 * @return void
	 */
	public function __construct() {
		$this->init();
	}

	/**
	 * Run action and filter hooks.
	 *
	 * @return void
	 */
	private function init() {
		add_action( 'wp_ajax_sisanu_image_replacement_from_url', array( $this, 'image_replacement_from_url' ) );
		add_action( 'wp_ajax_sisanu_image_replacement_from_upload', array( $this, 'image_replacement_from_upload' ) );
		add_action( 'wp_ajax_nopriv_sisanu_image_replacement_from_upload', array( $this, 'image_replacement_from_upload' ) );
		add_filter( 'admin_post_thumbnail_html', array( get_called_class(), 'image_replace_ajax_elements' ), 60, 3 );
		add_action( 'edit_form_after_title', array( get_called_class(), 'image_replace_ajax_elements_edit' ), 10 );
		add_action( 'admin_enqueue_scripts', array( get_called_class(), 'load_assets' ) );
		add_action( 'admin_notices', array( get_called_class(), 'plugin_admin_notices' ) );
		add_action( 'wp_ajax_srifu-plugin-deactivate-notice', array( get_called_class(), 'plugin_admin_notices_cleanup' ) );
		add_action( 'plugins_loaded', array( get_called_class(), 'plugin_ver_check' ) );
	}

	/**
	 * Load the plugin assets.
	 *
	 * @return void
	 */
	public static function load_assets() {
		wp_register_script(
			'serifu-js',
			plugins_url( '/assets/script.js', plugin_basename( __FILE__ ) ),
			array( 'jquery' ),
			SERIFU_PLUGIN_VERSION
		);
		wp_localize_script(
			'serifu-js',
			'serifuSettings',
			array(
				'ajaxUrl'      => admin_url( 'admin-ajax.php' ),
				'defaultType'  => 'upload', // Maybe configure in a new version of the plugin.
				'emptyUrl'     => esc_html__( 'You must specify an image URL.', 'srifu' ),
				'failedFetch'  => esc_html__( 'The image could not be fetched.', 'srifu' ),
				'emptyFile'    => esc_html__( 'You must select an image file.', 'srifu' ),
				'failedUpload' => esc_html__( 'The image could not be uploaded.', 'srifu' ),
				'noFile'       => esc_html__( 'Select a file.', 'srifu' ),
			)
		);
		wp_enqueue_script( 'serifu-js' );

		wp_enqueue_style(
			'serifu-style',
			plugins_url( '/assets/styles.css', plugin_basename( __FILE__ ) ),
			SERIFU_PLUGIN_VERSION
		);
	}

	/**
	 * Return the data mapped in the _wp_attachment_metadata of an attachemnt.
	 *
	 * @param integer $initial_image_id The attachment ID.
	 * @return array
	 */
	public static function get_initial_image_metadata( $initial_image_id ) {
		$metadata = maybe_unserialize( get_post_meta( $initial_image_id, '_wp_attachment_metadata', true ) );
		return $metadata;
	}

	/**
	 * Remove the files that were mapped in the _wp_attachment_metadata of an attachemnt.
	 *
	 * @param integer $initial_image_id The attachment ID.
	 * @param string  $basedir          The base dir.
	 * @return void
	 */
	public static function maybe_remove_files_from_metadata( $initial_image_id, $basedir ) {
		$original_folder = '';
		$removable = array();

		$metadata = self::get_initial_image_metadata( $initial_image_id );
		if ( ! empty( $metadata['file'] ) ) {
			$original_folder = dirname( $metadata['file'] );
			$removable['original'] = $basedir . '/' . $metadata['file'];
		}
		if ( ! empty( $metadata['sizes'] ) ) {
			$sizes = wp_list_pluck( $metadata['sizes'], 'file' );
			foreach ( $sizes as $size => $file ) {
				$removable[ $size ] = $basedir . '/' . $original_folder . '/' . $file;
			}
		}
		if ( ! empty( $removable ) ) {
			foreach ( $removable as $size => $file ) {
				wp_delete_file( $file );
			}
		}
	}

	/**
	 * Return the fail and default response.
	 *
	 * @return array
	 */
	public static function return_fail_response() {
		return array(
			'url'      => '',
			'new_info' => false,
			'old_info' => false,
			'changed'  => false,
			'srcset'   => '',
		);
	}

	/**
	 * Return the processed guid, new and old metadata, if the image from the specified URL
	 * could be fetched and saved in the uploads directory as the attachment file.
	 *
	 * @param integer $initial_image_id The attachment ID.
	 * @param string  $image_url        The URL of the image used as the replacement.
	 * @return boolean|array
	 */
	public static function maybe_replace_image_from_url( $initial_image_id, $image_url ) {
		if ( ! empty( $initial_image_id ) && ! empty( $image_url ) ) {
			$new_file_content = '';

			// Let's fetch the remote image.
			$response = wp_remote_get( $image_url );
			$code = wp_remote_retrieve_response_code( $response );
			if ( 200 === $code ) {
				// Seems that we got a successful response from the remore URL.
				$content_type = wp_remote_retrieve_header( $response, 'content-type' );
				if ( ! empty( $content_type ) && substr_count( $content_type, 'image/' ) ) {
					// Seems that the content type is an image, let's get the body as the file content.
					$new_file_content = wp_remote_retrieve_body( $response );
				}
			}

			if ( ! empty( $new_file_content ) ) {
				return self::maybe_replace_image_from_file_content( $initial_image_id, $new_file_content );
			}
		}

		return self::return_fail_response();
	}

	/**
	 * Return the processed guid, new and old metadata, if the image from the specified URL
	 * could be fetched and saved in the uploads directory as the attachment file.
	 *
	 * @param integer $initial_image_id The attachment ID.
	 * @param string  $file             The URL of the image used as the replacement.
	 * @return boolean|array
	 */
	public static function maybe_replace_image_from_upload( $initial_image_id, $file ) {
		if ( ! empty( $initial_image_id ) && ! empty( $file ) ) {
			$new_file_content = '';

			$allowed_types = array(
				'image/gif',
				'image/jpg',
				'image/jpeg',
				'image/png',
			);

			// Let's check that the file was upoloaded.
			if ( empty( $file['erorr'] ) && ! empty( $file['tmp_name'] ) && ! empty( $file['type'] ) && in_array( $file['type'], $allowed_types ) ) {
				$new_file_content = file_get_contents( $file['tmp_name'] );
			}

			if ( ! empty( ! empty( $file['tmp_name'] ) ) ) {
				@unlink( $file['tmp_name'] );
			}
			if ( ! empty( $new_file_content ) ) {
				return self::maybe_replace_image_from_file_content( $initial_image_id, $new_file_content );
			}
		}

		return self::return_fail_response();
	}

	/**
	 * Return the processed guid, new and old metadata, if the image from the specified URL
	 * could be fetched and saved in the uploads directory as the attachment file.
	 *
	 * @param integer $initial_image_id The attachment ID.
	 * @param string  $new_file_content The URL of the image used as the replacement.
	 * @return boolean|array
	 */
	public static function maybe_replace_image_from_file_content( $initial_image_id, $new_file_content ) {
		if ( ! empty( $initial_image_id ) && ! empty( $new_file_content ) ) {
			$upload_dir = wp_upload_dir();
			$basedir = $upload_dir['basedir'];

			$post = get_post( $initial_image_id );
			$meta = get_post_meta( $initial_image_id, '_wp_attached_file', true );
			if ( ! empty( $meta ) ) {
				$image_path = $basedir . '/' . $meta;
			} else {
				$image_path = $upload_dir['path'] . '/' . basename( $image_url );
			}

			if ( ! empty( $image_path ) && file_exists( $image_path ) ) {
				// Cleanup existing files.
				self::maybe_remove_files_from_metadata( $initial_image_id, $basedir );
			}

			if ( @file_put_contents( $image_path, $new_file_content ) ) {
				// @TODO - maybe use wp_upload_bits when that function allows to specify the descrination, or we will have WP native hooks.
				$file_info = getimagesize( $image_path );

				if ( ! function_exists( 'wp_generate_attachment_metadata' ) ) {
					require_once( ABSPATH . 'wp-admin/includes/image.php' );
				}

				// Get the initial image info.
				$old_data = self::get_initial_image_metadata( $initial_image_id );

				$new_data = wp_generate_attachment_metadata( $initial_image_id, $image_path );
				if ( ! empty( $new_data['file'] ) ) {
					// This means that the metadata was generated for the new image.
					// Allow to filter and maybe override the new metadata.
					$new_data = apply_filters( 'sisanu_replace_image_before_update_attachment_info', $new_data, $old_data );
					wp_update_attachment_metadata( $initial_image_id, $new_data );
				}

				$artdata = array(
					'ID' => (int) $initial_image_id,
					'guid' => wp_get_attachment_url( (int) $initial_image_id ),
					'post_mime_type' => $file_info['mime'],
				);

				$artdata = apply_filters( 'sisanu_replace_image_before_update_attachment_post', $artdata );
				wp_update_post( $artdata );

				$srcset = self::make_srcset( $new_data, $old_data );

				// Return the image URL.
				return array(
					'url'      => $artdata['guid'],
					'new_info' => $new_data,
					'old_info' => $old_data,
					'changed'  => true,
					'srcset'   => $srcset,
				);
			}
		}

		return self::return_fail_response();
	}

	/**
	 * Generate a srcset attribute value from the image metadata.
	 *
	 * @param array $new_data An image metadata array.
	 * @return string
	 */
	public static function make_srcset( $new_data ) {
		$srcset = '';
		$upload_dir = wp_upload_dir();
		$baseurl = trailingslashit( $upload_dir['baseurl'] );
		$folder = ( ! empty( $new_data['file'] ) ) ? dirname( $new_data['file'] ) : '';
		if ( ! empty( $new_data['sizes'] ) ) {
			foreach ( $new_data['sizes'] as $size => $file_info ) {
				$srcset .= ( '' === $srcset ) ? '' : ', ';
				$srcset .= trailingslashit( $baseurl . $folder ) . $file_info['file'] . '?v=' . time() . ' ' . $file_info['width'] . 'w';
			}
		}

		return $srcset;
	}

	/**
	 * Append the plugin elements used for the replacement of the image in the edit thumbnail code.
	 *
	 * @param  string  $content      The post thumbnail html.
	 * @param  integer $post_id      The current post ID.
	 * @param  integer $thumbnail_id The thumbnail ID.
	 * @return string
	 */
	public static function image_replace_ajax_elements( $content, $post_id = 0, $thumbnail_id = 0 ) {
		$attachment_id = 0;
		$display = false;
		if ( ! empty( $thumbnail_id ) ) {
			$attachment_id = $thumbnail_id;
		} else {
			if ( is_object( $post_id ) ) {
				$post_id = $post_id->ID;
			}
			$post = get_post( $post_id );
			if ( ! empty( $post->post_type ) && 'attachment' === $post->post_type ) {
				$attachment_id = $post_id;
			}
		}
		if ( ! empty( $attachment_id ) ) {
			$button = self::refresh_from_url_input( $attachment_id, '' );
			$content = $content . $button;
		}

		return $content;
	}

	/**
	 * Outputs the plugin elements used for the replacement of the image in the edit form after title.
	 *
	 * @param object $post The current post.
	 * @return void
	 */
	public static function image_replace_ajax_elements_edit( $post ) {
		if ( is_object( $post ) && ! empty( $post->post_type ) && 'attachment' === $post->post_type ) {
			$allowed_tags = array(
				'a' => array(
					'href'  => array(),
					'title' => array(),
					'class' => array(),
					'data-attachment-id' => array(),
				),
				'div' => array(
					'id' => array(),
					'class' => array(),
				),
				'table' => array(
					'class' => array(),
					'width' => array(),
					'cellspacing' => array(),
					'cellpadding' => array(),
				),
				'tr' => array(),
				'td' => array(
					'width' => array(),
					'align' => array(),
					'class' => array(),
				),
				'span' => array(
					'class' => array(),
				),
				'img' => array(
					'src' => array(),
					'class' => array(),
				),
				'hr' => array(),
				'em' => array(),
				'b' => array(),
				'p' => array(),
				'strong' => array(),
				'br' => array(),
				'label' => array(),
				'input' => array(
					'class' => array(),
					'type' => array(),
					'placeholder' => array(),
					'checked' => array(),
					'name' => array(),
					'id' => array(),
					'value' => array(),
					'size' => array(),
				),
			);
			echo wp_kses( self::refresh_from_url_input( $post->ID, '<div class="inside">', '</div>' ), $allowed_tags );
		}
	}

	/**
	 * Returns the markup for the elements used by the image replace functionality.
	 *
	 * @param  integer $attachment_id The attachment ID.
	 * @param  string  $before        Markup before the elements.
	 * @param  string  $after         Markup after the elements.
	 * @return string
	 */
	public static function refresh_from_url_input( $attachment_id, $before = '', $after = '' ) {
		$button = '
		<p>
			<label for="js-serifu-image-fetch-wrap-' . (int) $attachment_id . '">
				<strong>' . esc_html__( 'EASY REPLACE IMAGE', 'srifu' ) . '</strong>
			</label>
			<div class="js-serifu-image-fetch-wrap" id="js-serifu-image-fetch-wrap-' . (int) $attachment_id . '">
				' . wp_kses_post( $before ) . '
				<img src="' . esc_url( plugins_url( '/assets/replace.png', __FILE__ ) ) . '" class="js-serifu-image-icon">
				<label><input type="radio" name="serifu_replace_type" id="serifu_replace_type_upload" value="upload">' . esc_attr__( 'Upload', 'srifu' ) . '</label>
				<label><input type="radio" name="serifu_replace_type" id="serifu_replace_type_url" value="url" checked="checked">' . esc_attr__( 'Get from URL', 'srifu' ) . '</label>
				<br>
				<div class="js-serifu-replace-type-upload">
					' . esc_attr__( 'Replace* the attachment file with the image you upload', 'srifu' ) . ':
					<table cellspacing="0" cellpadding="1" class="fixed" width="100%">
						<tr>
							<td width="32" class="file-wrap upload"><span class="dashicons dashicons-upload"></span></td>
							<td width="100%" class="file-wrap">
								<div class="js-serifu-image-upload-label">' . esc_attr__( 'Select a file.', 'srifu' ) . '</div>
								<input id="js-serifu-image-upload-file-' . (int) $attachment_id . '" type="file" class="js-serifu-image-upload" size="200">
							</td>
							<td width="100" align="left"><a class="button js-serifu-image-upload" data-attachment-id="' . (int) $attachment_id . '">' . esc_html__( 'Upload', 'srifu' ) . '</a></td>
						</tr>
					</table>
					<em>' . esc_attr__( '* If the image you specify can be uploaded to your site uploads folder, it will replace the current file.', 'srifu' ) . '</em>
				</div>
				<div class="js-serifu-replace-type-url off">
					' . esc_attr__( 'Replace* the attachment file with the image from the specified URL', 'srifu' ) . ':
					<table cellspacing="0" cellpadding="1" class="fixed" width="100%">
						<tr>
							<td width="32" class="file-wrap download"><span class="dashicons dashicons-download"></span></td>
							<td width="100%" class="file-wrap"><input type="text" class="js-serifu-image-fetch" placeholder="' . esc_attr__( 'Image URL', 'srifu' ) . '"></td>
							<td width="100" align="left"><a class="button js-serifu-image-fetch" data-attachment-id="' . (int) $attachment_id . '">' . esc_html__( 'Download', 'srifu' ) . '</a></td>
						</tr>
					</table>
					<em>' . esc_attr__( '* If the image you specify can be downloaded to your site uploads folder, it will replace the current file.', 'srifu' ) . '</em>
				</div>
				<span class="js-serifu-image-processed-response"></span>
				' . wp_kses_post( $after ) . '
			</div>
		</p>
		';
		return $button;
	}

	/**
	 * AJAX handler for replecing the file of a specified attachment from a URL.
	 *
	 * @access public
	 * @static
	 * @return void
	 */
	public function image_replacement_from_url() {

		$old_image_id = filter_input( INPUT_POST, 'old_image_id', FILTER_VALIDATE_INT );
		$new_image_url = filter_input( INPUT_POST, 'new_image_url', FILTER_SANITIZE_URL );
		if ( ! empty( $old_image_id ) && ! empty( $new_image_url ) ) {
			$response = self::maybe_replace_image_from_url( $old_image_id, $new_image_url );
			if ( ! empty( $response ) && ! empty( $response['url'] ) ) {
				$response['url'] = $response['url'] . '?v=' . current_time( 'timestamp' );
				echo wp_json_encode( $response );
			}
		}
		die();
	}

	/**
	 * AJAX handler for replecing the file of a specified attachment from an upload.
	 *
	 * @access public
	 * @static
	 * @return void
	 */
	public function image_replacement_from_upload() {
		$old_image_id = filter_input( INPUT_POST, 'old_image_id', FILTER_VALIDATE_INT );
		if ( ! empty( $old_image_id ) && ! empty( $_FILES[0] ) ) {
			$response = self::maybe_replace_image_from_upload( $old_image_id, $_FILES[0] );
			if ( ! empty( $response ) && ! empty( $response['url'] ) ) {
				$response['url'] = $response['url'] . '?v=' . current_time( 'timestamp' );
				echo wp_json_encode( $response );
			}
		}
		die();
	}

	/**
	 * The actions to be executed when the plugin is activated.
	 *
	 * @return void
	 */
	public static function activate_plugin() {
		set_transient( self::PLUGIN_TRANSIENT, true );
	}

	/**
	 * The actions to be executed when the plugin is deactivated.
	 *
	 * @return void
	 */
	public static function deactivate_plugin() {
		self::plugin_admin_notices_cleanup( false );
	}

	/**
	 * The actions to be executed when the plugin is updated.
	 *
	 * @return void
	 */
	public static function plugin_ver_check() {
		$opt = str_replace( '-', '_', self::PLUGIN_TRANSIENT ) . '_db_ver';
		$dbv = get_option( $opt, 0 );
		if ( SERIFU_PLUGIN_VERSION !== (float) $dbv ) {
			update_option( $opt, SERIFU_PLUGIN_VERSION );
			self::activate_plugin();
		}
	}

	/**
	 * Execute notices cleanup.
	 *
	 * @param  boolean $ajax Is AJAX call.
	 * @return void
	 */
	public static function plugin_admin_notices_cleanup( $ajax = true ) {
		// Delete transient, only display this notice once.
		delete_transient( self::PLUGIN_TRANSIENT );

		if ( true === $ajax ) {
			// No need to continue.
			wp_die();
		}
	}

	/**
	 * Admin notices.
	 *
	 * @return void
	 */
	public static function plugin_admin_notices() {
		$maybe_trans = get_transient( self::PLUGIN_TRANSIENT );
		if ( ! empty( $maybe_trans ) ) {
			?>
			<style>.notice.<?php echo esc_attr( self::PLUGIN_TRANSIENT ); ?>{background:rgba(60,142,131,0.2);border-left-color:rgb(60,142,131)}.notice.<?php echo esc_attr( self::PLUGIN_TRANSIENT ); ?> img{max-width: 100%}</style>
			<script>(function($) { $(document).ready(function() { var $notice = $('.notice.<?php echo esc_attr( self::PLUGIN_TRANSIENT ); ?>'); var $button = $notice.find('.notice-dismiss'); $notice.unbind('click'); $button.unbind('click'); $notice.on('click', '.notice-dismiss', function(e) { $.get( $notice.data('dismissurl') ); }); }); })(jQuery);</script>

			<div class="updated notice is-dismissible <?php echo esc_attr( self::PLUGIN_TRANSIENT ); ?>"
				data-dismissurl="<?php echo esc_url( admin_url( 'admin-ajax.php?action=srifu-plugin-deactivate-notice' ) ); ?>">
				<p>
					<?php
					echo wp_kses_post( sprintf(
						// Translators: %1$s - image URL, %2$s - icon URL, %3$s - donate URL, %4$s - link style, %5$s - icon style, %6$s - rating.
						__( '<a href="%3$s" target="_blank"%4$s><img src="%1$s"></a><img src="%2$s"%5$s> <h3>Thank you for activating the plugin Easy Replace Image!</h3>If you find the plugin useful and would like to support my work, please consider making a <a href="%3$s" target="_blank">donation</a>. It would make me very happy if you would leave a %6$s rating.', 'srifu' ) . ' ' . __( 'A huge thanks in advance!', 'srifu' ),
						esc_url( plugin_dir_url( __FILE__ ) . '/assets/images/buy-me-a-coffee.png' ),
						esc_url( plugin_dir_url( __FILE__ ) . '/assets/images/icon-128x128.png' ),
						'https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=JJA37EHZXWUTJ&item_name=Support for development and maintenance (' . self::PLUGIN_NAME . ')',
						' style="float:right; margin:20px"',
						' style="float:left; margin-right:20px; margin-top:10px; width:86px"',
						'<a href="' . self::PLUGIN_SUPPORT_URL . 'reviews/?rate=5#new-post" class="rating" target="_blank" title="' . esc_attr( 'A huge thanks in advance!', 'srifu' ) . '">★★★★★</a>'
					) );
					?>
					<div class="clear"></div>
				</p>
			</div>
			<?php
		}
	}
}

// Instantiate the class.
$srifu = Sisanu_Easy_Replace_Image::get_instance();

register_activation_hook( __FILE__, array( $srifu, 'activate_plugin' ) );
register_deactivation_hook( __FILE__, array( $srifu, 'deactivate_plugin' ) );
